"""
Function that ignite from main_window, Dialogs to read data files for data,
channels, datatype
"""

import os
import json
import re
from pathlib import Path
from typing import List, Set, Optional, Dict, Tuple

from PySide2.QtCore import QEventLoop, Qt
from PySide2.QtGui import QCursor
from PySide2.QtWidgets import QTextBrowser, QApplication
from obspy.io import reftek

from sohstationviewer.model.mseed_data.record_reader import RecordReader \
    as MSeedRecordReader
from sohstationviewer.model.mseed_data.record_reader_helper import \
    MSeedReadError
from sohstationviewer.model.mseed_data.mseed_reader import \
    move_to_next_record

from sohstationviewer.database.extract_data import get_signature_channels
from sohstationviewer.model.handling_data import (
    read_mseed_chanids_from_headers)

from sohstationviewer.controller.util import (
    validate_file, display_tracking_info
)

from sohstationviewer.view.util.enums import LogType


def read_mseed_channels(tracking_box: QTextBrowser, list_of_dir: List[str],
                        on_unittest: bool = False
                        ) -> Set[str]:
    """
    Scan available for SOH channels (to be used in channel preferences dialog).
        Since channels for RT130 is hard code, this function won't be applied
        for it.
    Note that Mass position channels are excluded because the default
        include_mp123zne and include_mp456uvw for MSeed are False
    :param tracking_box: widget to display tracking info
    :param list_of_dir: list of directories selected by users
    :param on_unittest: flag to avoid cursor code to display tracking_info
    :return data_object.channels: set of channels present in listofDir
    """
    soh_chan_ids = set()
    mass_pos_chan_ids = set()
    wf_chan_ids = set()
    spr_gr_1_chan_ids = set()
    count = 0
    if not on_unittest:
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
    display_tracking_info(
        tracking_box, "Start reading channel's header", LogType.INFO)
    for d in list_of_dir:
        for path, sub_dirs, files in os.walk(d):
            if 'System Volume Information' in path:
                continue
            for file_name in files:
                if not on_unittest:
                    QApplication.processEvents(
                        QEventLoop.ExcludeUserInputEvents
                    )
                path2file = Path(path).joinpath(file_name)
                if not validate_file(path2file, file_name):
                    continue
                count += 1
                if count % 10 == 0:
                    display_tracking_info(
                        tracking_box, f'Read {count} file headers/ SOH files',
                        LogType.INFO)

                ret = read_mseed_chanids_from_headers(path2file, file_name)
                soh_chan_ids.update(ret[0])
                mass_pos_chan_ids.update(ret[1])
                wf_chan_ids.update(ret[2])
                spr_gr_1_chan_ids.update(ret[3])
    if not on_unittest:
        QApplication.restoreOverrideCursor()
    return sorted(list(soh_chan_ids)), sorted(list(mass_pos_chan_ids)), \
        sorted(list(wf_chan_ids)), sorted(list(spr_gr_1_chan_ids))


def detect_data_type(list_of_dir: List[str]) -> Optional[str]:
    """
    Detect data type for the given directories using get_data_type_from_file
    :param list_of_dir: list of directories selected by users
    :return:
        + if there are more than one data types detected,
            return None with a warning message
        + if only Unknown data type detected,
            return None with a warning message
        + if data type found, return data_type,
    """
    sign_chan_data_type_dict = get_signature_channels()

    dir_data_type_dict = {}
    is_multiplex = None
    for d in list_of_dir:
        data_type = "Unknown"
        for path, subdirs, files in os.walk(d):
            for file_name in files:
                path2file = Path(path).joinpath(file_name)
                if not validate_file(path2file, file_name):
                    continue
                ret = get_data_type_from_file(path2file,
                                              sign_chan_data_type_dict,
                                              is_multiplex)
                if ret is not None:
                    d_type, is_multiplex = ret
                    if d_type is not None:
                        data_type = d_type
                        break
            if data_type != "Unknown":
                break

        if is_multiplex is None:
            raise Exception("No channel found for the data set")

        if data_type == "Unknown":
            dir_data_type_dict[d] = "Unknown"
        else:
            dir_data_type_dict[d] = data_type
    data_type_list = list(set(dir_data_type_dict.values()))
    if len(data_type_list) > 1:
        dir_data_type_str = json.dumps(dir_data_type_dict)
        dir_data_type_str = re.sub(r'\{|\}|"', '', dir_data_type_str)
        dir_data_type_str = re.sub(r'], ', ']\n', dir_data_type_str)
        msg = (f"There are more than one types of data detected:\n"
               f"{dir_data_type_str}\n\n"
               f"Please have only data that related to each other.")
        raise Exception(msg)

    elif data_type_list == ['Unknown']:
        msg = ("There are no known data detected.\n\n"
               "Do you want to cancel to select different folder(s)\n"
               "Or continue to read any available mseed file?")
        raise Exception(msg)
    return data_type_list[0], is_multiplex


def get_data_type_from_file(
        path2file: Path,
        sign_chan_data_type_dict: Dict[str, str],
        is_multiplex: bool = None
) -> Optional[Tuple[Optional[str], bool]]:
    """
    + Exclude files for waveform data to improve performance
    + Loop through each record for file
        If MSeedRecordReader gives Error; check if the file is RT130, report
            data_type is RT130 or else, return to continue checking on another
            file.
        If there're more than one channels in a file, this file is multiplex.
        If found signature channel, report the data_type of the file.
    :param path2file: absolute path to processed file
    :param sign_chan_data_type_dict: dict of unique chan for data
        type
    :param is_multiplex: if the file is multiplex
    :return: detected data type, channel from which data type is detected
    """
    wf_chan_posibilities = ['FH', 'FN',  # ≥ 1000 to < 5000
                            'GH', 'GL',  # ≥ 1000 to < 5000
                            'DH', 'DL',  # ≥ 250 to < 1000
                            'CH', 'CN',  # ≥ 250 to < 1000
                            'EH', 'EL', 'EP',  # ≥ 80
                            'SH', 'SL', 'SP',  # ≥ 10 to < 80
                            'HH', 'HN',  # ≥ 80
                            'BH', 'BN',  # ≥ 10 to < 80
                            'MH', 'MN', 'MP', 'ML',
                            'LH', 'LL', 'LP', 'LN',
                            'VP', 'VL', 'VL', 'VH',
                            'UN', 'UP', 'UL', 'UH']

    if any(x in path2file.name for x in wf_chan_posibilities):
        # Skip checking waveform files which aren't signature channels
        return None, False

    file = open(path2file, 'rb')
    chans_in_stream = set()
    data_type = None
    while 1:
        is_eof = (file.read(1) == b'')
        if is_eof:
            break
        file.seek(-1, 1)
        current_record_start = file.tell()
        try:
            record = MSeedRecordReader(file)
        except MSeedReadError:
            file.close()
            if reftek.core._is_reftek130(path2file):
                return 'RT130', False
            return

        chan = record.record_metadata.channel
        if is_multiplex is None:
            chans_in_stream.add(chan)
            if len(chans_in_stream) > 1:
                is_multiplex = True
        if chan in sign_chan_data_type_dict.keys():
            data_type = sign_chan_data_type_dict[chan]
            if is_multiplex:
                file.close()
                return data_type, is_multiplex
        move_to_next_record(file, current_record_start, record)
    file.close()
    is_multiplex = True if len(chans_in_stream) > 1 else False
    return data_type, is_multiplex
